package edu.ucf.cs.courses.cop4610l.beans.crypto;

/**
 * @author Mathew B. Lowery
 * @version 1.0
 */

import java.io.*;
import java.util.*;
import javax.crypto.*;
import java.security.*;
import java.security.cert.*;

/**
 * A JavaBean that encapsulates common encryption functions.  It takes as an
 * argument an <code>InputStream</code> and returns (via an event) another
 * <code>InputStream</code> that can be consumed by another object.
 */
public class Crypto implements Serializable {
  /**
   * The symmetric algorithm for this bean to use (default is "DES").
   */
  private String symmetricCipherName = "DES";
  /**
   * Filename of the keystore of type "JCEKS" which holds all keys generated
   * by any beans of this type.
   */
  private static final String KEYSTORE_FILENAME = "cop4610L-keystore";
  /**
   * Password for keystore in <code>KEYSTORE_FILENAME</code>.
   */
  private static final String KEYSTORE_PASSWORD = "passwd";
  /**
   * Password for individual keys in the keystore.  (All key passwords are the
   * same.)
   */
  private static final String KEY_PASSWORD = "passwd";
  /**
   * Prefix for all keys created by beans of this type.  This allows to bean
   * to only use keys created by it even though others might exist for a given
   * algorithm.
   */
  private static final String KEY_PREFIX = "cop4610L";
  /**
   * Listeners of <code>TransformCompletedEvent</code>s.
   */
  private Set listeners = new HashSet();
  /**
   * The stream that is consumed by listeners when a
   * <code>TransformCompletedEvent</code> is fired.
   */
  private InputStream symmetricTransformedStream;

  public Crypto() {
  } // constructor

  /**
   * There is only a getter method for this property since it is read-only.  It
   * should be called only after hearing a <code>TransformCompletedEvent</code>.
   * @return the transformed stream
   */
  public InputStream getSymmetricTransformedStream() {
    return symmetricTransformedStream;
  }

  public String getSymmetricCipherName() {
    return symmetricCipherName;
  } // getSymmetricCipherName

  public void setSymmetricCipherName(String symmetricCipherName) {
    this.symmetricCipherName = symmetricCipherName;
  } // setSymmetricCipherName

  /**
   * Encrypts contents of <code>in</code> using the algorithm specified by
   * <code>symmetricCipherName</code>.  A <code>TransformCompletedEvent</code>
   * is fired when done.
   * @param in the stream to encrypt
   */
  public void doSymmetricEncrypt(InputStream in) {
    doSymmetricTransform(in, Cipher.ENCRYPT_MODE);
  } // doSymmetricEncrypt

  /**
   * Decrypts contents of <code>in</code> using the algorithm specified by
   * <code>symmetricCipherName</code>.  A <code>TransformCompletedEvent</code>
   * is fired when done.
   * @param in the stream to decrypt
   */
  public void doSymmetricDecrypt(InputStream in) {
    doSymmetricTransform(in, Cipher.DECRYPT_MODE);
  } // doSymmetricDecrypt

  /**
   * This method does the following:
   *   Loads keystore if it exists on disk.  Otherwise it creates a new one.
   *   If the key with the algorithm given by <code>cipherName</code> exists,
   *   it returns it. Otherwise it creates a new key with the given algorithm,
   *   stores it in the keystore, then saves the updated keystore.
   * @param cipherName algorithm of key to look for
   * @return existing key or new one if no existing key
   */
  private Key findKey(String cipherName) throws Exception {
    // used "jceks" since "jks" didn't work
    //   see Java Developer Connection forum 9, thread 14793
    KeyStore keyStore = KeyStore.getInstance("JCEKS");
    File keyStoreFile = new File(KEYSTORE_FILENAME);
    if (keyStoreFile.exists()) {
      FileInputStream fin = new FileInputStream(keyStoreFile);
      // load keystore into memory from file
      keyStore.load(fin, KEYSTORE_PASSWORD.toCharArray());
      fin.close();
    }
    else {
      // create new keystore in memory
      keyStore.load(null, KEYSTORE_PASSWORD.toCharArray());
    }

    // find the key for cipherName if it exists
    Enumeration aliases = keyStore.aliases();
    Key key = null;
    while (aliases.hasMoreElements()) {
      String currentAlias = (String) aliases.nextElement();
      // when keys are put in keystore the case gets converted to all lower
      if (currentAlias.startsWith(KEY_PREFIX.toLowerCase())) {
        Key currentKey = keyStore.getKey(currentAlias,
                                         KEY_PASSWORD.toCharArray());
        if (currentKey.getAlgorithm().equals(cipherName)) {
          key = currentKey;
          break;
        }
      }
    }

    if (key == null) {
      // key does not exist in keystore; create new key, store it, and save
      //   new keystore

      // get key generator instance
      KeyGenerator keyGenerator = null;
      try {
        keyGenerator = KeyGenerator.getInstance(cipherName);
      }
      catch (NoSuchAlgorithmException e) {
        e.printStackTrace();
      }
      // generate key
      key = keyGenerator.generateKey();
      // store key in keystore
      String newKeyName = KEY_PREFIX + new Integer(keyStore.size()).toString();
      keyStore.setKeyEntry(newKeyName, key, KEY_PASSWORD.toCharArray(),
                           null);
      // save keystore to file
      FileOutputStream fout = new FileOutputStream(KEYSTORE_FILENAME);
      keyStore.store(fout, KEYSTORE_PASSWORD.toCharArray());
      fout.close();
    }
    return key;
  } // findKey

  /**
   * Used by both <code>doSymmmetricEncrypt</code> and
   * <code>doSymmetricDecrypt</code>.  The only difference between the two
   * methods is the <code>Cipher</code> mode.
   * @param in
   * @param cipherMode
   */
  private void doSymmetricTransform(InputStream in, int cipherMode) {
    // create pipes
    PipedOutputStream pipeOut = null;
    PipedInputStream pipeIn = null;
    try {
      // whatever is written to pipeOut can be read from pipeIn
      pipeOut = new PipedOutputStream();
      pipeIn = new PipedInputStream(pipeOut);
    }
    catch (IOException e) {
      e.printStackTrace();
    }

    // create a cipher object which represents a specific algorithm
    Cipher cipher = null;
    try {
      cipher = Cipher.getInstance(symmetricCipherName);
    }
    catch (NoSuchPaddingException e) {
      e.printStackTrace();
    }
    catch (NoSuchAlgorithmException e) {
      e.printStackTrace();
    }

    Key key = null;
    try {
      key = findKey(symmetricCipherName);
    }
    catch (Exception e) {
      e.printStackTrace();
    }

    // initialize cipher to encrypt or decrypt
    try {
      cipher.init(cipherMode, key);
    }
    catch (InvalidKeyException e) {
      e.printStackTrace();
    }

    // consume input stream, encrypt, and write encrypted data to output stream
    CipherInputStream cin = new CipherInputStream(in, cipher);
    byte[] b = new byte[8];
    try {
      int i = cin.read(b);
      while (i != -1) {
        pipeOut.write(b, 0, i);
        i = cin.read(b);
      } // while
      pipeOut.close();
    }
    catch (IOException e) {
      e.printStackTrace();
    }
    // set property so listeners can read it
    symmetricTransformedStream = pipeIn;
    fireTransformCompleted(new TransformCompletedEvent(this, pipeIn));
  } // doSymmetricTransform

  /**
   * This bean currently can only support one listener.
   * @param listener
   * @throws TooManyListenersException if more than one listener attempts to add
   */
  public void addTransformCompletedListener(
                                      TransformCompletedListener listener)
                                      throws TooManyListenersException {
    synchronized (listeners) {
      // no listeners yet; add it
      if (listeners.isEmpty()) {
        listeners.add(listener);
      }
      else {
        if (listeners.contains(listener)) {
          // OK; just return
          return;
        }
        else {
          // trying to add new listener when another exists already
          throw new TooManyListenersException(
                "only one listener allowed per instance of this class");
        }
      }
    } // sychronized
  } // addTransformedCompletedListener

  public void removeTransformCompletedListener(
                                      TransformCompletedListener listener) {
    synchronized (listeners) {
      listeners.remove(listener);
    }
  }

  private void fireTransformCompleted(TransformCompletedEvent e) {
    Iterator iterator;
    synchronized (listeners) {
      iterator = new HashSet(listeners).iterator();
    }
    while (iterator.hasNext()) {
      ((TransformCompletedListener) iterator.next()).transformCompleted(e);
    }
  } // fireTransformCompleted
} // Crypto