package edu.ucf.cs.courses.cop4610l.beans.streams;

/**
 * @author Mathew B. Lowery
 * @version 1.0
 */

import java.io.*;
import java.util.*;

/**
 * A JavaBean that puts the contents of a <code>String</code> into a
 * <code>InputStream</code>.  It takes as an argument a string and returns
 * (via an event) a stream that can be consumed by other objects.
 */
public class StringToStream implements Serializable {
  private Set listeners = new HashSet();
  private InputStream stream;

  public StringToStream() {
  }

  /**
   * There is only a getter method for this property since it is read-only.  It
   * should be called only after hearing a <code>StreamConstructedEvent</code>.
   * @return the newly constructed stream
   */
  public InputStream getStream() {
    return stream;
  }

  /**
   * This bean currently can only support one listener.
   * @param listener
   * @throws TooManyListenersException if more than one listener attempts to add
   */
  public void addStreamConstructedListener(StreamConstructedListener listener)
                                           throws TooManyListenersException {
    synchronized (listeners) {
      // no listeners yet; add it
      if (listeners.isEmpty()) {
        listeners.add(listener);
      }
      else {
        if (listeners.contains(listener)) {
          // OK; just return
          return;
        }
        else {
          // trying to add new listener when another exists already
          throw new TooManyListenersException(
                "only one listener allowed per instance of this class");
        }
      }
    } // sychronized
  }

  public void removeStreamConstructedListener(
                                      StreamConstructedListener listener) {
    synchronized (listeners) {
      listeners.remove(listener);
    }
  }

  private void fireStreamConstructed(StreamConstructedEvent e) {
    Iterator iterator;
    synchronized (listeners) {
      iterator = new HashSet(listeners).iterator();
    }
    while (iterator.hasNext()) {
      ((StreamConstructedListener) iterator.next()).streamConstructed(e);
    }
  }

  /**
   * Puts the contents of <code>s</code> into an <code>InputStream</code> for
   * consumption by another object.
   * @param s the string produced from the contents of the stream
   */
  public void constructStream(String s) {
    PipedOutputStream pipeOut = null;
    PipedInputStream pipeIn = null;
    try {
      // whatever is written to pipeOut can be read from pipeIn
      pipeOut = new PipedOutputStream();
      pipeIn = new PipedInputStream(pipeOut);
      pipeOut.write(s.getBytes());
      pipeOut.close();
    }
    catch (IOException e) {
      e.printStackTrace();
    }
    // set property so listeners can read it
    stream = pipeIn;
    fireStreamConstructed(new StreamConstructedEvent(this, pipeIn));
  }
} // StringToStream